--
--	Wholly
--	Written by scott@mithrandir.com
--
--	This was inspired by my need to replace EveryQuest with something that was more accurate.  The pins
--	was totally inspired by QuestHubber which I found when looking for a replacement for EveryQuest.  I
--	initially made a port of QuestHubber to QuestHubberGrail to make use of the Grail database, but now
--	I have integrated that work into this addon.  Many thanks for all the work put into QuestHubber.  I
--	was inspired to add a quest breadcrumb area from seeing one in Quest Completist.
--
--	Version History
--		001	Initial version.
--		002	Support for displaysDungeonQuests now works properly.
--			Added the ability for the panel tooltip to display questgivers.
--			Added the ability to click a quest in the panel to create a TomTom waypoint.
--			Map pins are only displayed for the proper dungeon level of the map.
--		003	Added a panel button to switch to the current zone.
--			Changed the Close button into a Sort button that switches between three different modes:
--				1. Alphabetical  2. Quest level, then alphabetical  3. Quest status, then alphabetical
--			Made map pins only have one pin per NPC, indicating the "best" color possible.
--			The entire zone dropdown button on the quest log panel now can be clicked to change zones.
--			Corrected a problem where someone running without LibStub would get a LUA error.
--			Corrected a localization issue.
--		004	Added the ability to display a basic quest prerequisite chain in the quest panel tooltip, requiring Grail 014 or later.
--			Added the ability to right-click a "prerequisite" quest in the panel to put a TomTom arrow to the first prerequisite quest.
--			Added Dungeons and Other menu items for map areas in the quest log panel.
--			The last-used sort preference is stored on a per-character basis.
--		005	Corrected the fix introduced in version 003 putting the LibDataBroker icon back in place.
--			Corrected a problem where the quest log tooltip would have an error if the questgiver was in a dungeon appearing in the zone.
--			Added the ability for the quest log tooltip to show that the questgiver should be killed (like the map pin tooltip).
--			The problem where map pins would not live update unless the quest log panel was opened has been fixed as long
--				as the Wholly check button appears on the map.
--			Added the ability to show quest breadcrumb information when the Quest Frame opens, showing a tooltip of breadcrumb
--				quests when the mouse enters the "breadcrumb area", and putting TomTom waypoints when clicking in it.
--		006	Added the new quest group "World Events" which has holiday quests in it, requiring Grail 015.
--			Added a tooltip to the Wholly check button on the map that indicates how many quests of each type are in the map.
--			Added a tooltip to the LibDataBroker icon that shows the quest log panel "map" selection and the quest count/type.
--			Added a tooltip to the quest log panel Zone button that shows the quest count/type.
--			Corrected the problem where the quest log panel and map pins were not live updating when quest givers inside dungeons checked.
--			Corrected the problem where an NPC that drops items that starts more than one quest does not display the information properly
--				in its tooltip.
--			Made it so the open World Map can be updated when crossing into a new zone.
--		007	Added the ability to show whether quests in the quest log are completed or failed.
--			Made it so right-clicking an "in log" quest will put in TomTom waypoints for the turn in locations, which requires Grail 016
--				for proper functioning since Grail 015 had a bug.
--			Made the strings for the preferences color quest information like it appears in the UI.
--			Made it so alt-clicking a log in the Wholly quest log selects the NPC that gives the quest or for the case of "in log" quests
--				the one to which the quest is turned in.
--		008	Split out Dungeons into dungeons in different continents, requiring Grail version 017.
--			Corrected a misspelling of the global game tooltip name.
--		009	Added localization for ptBR in anticipation of the Brazilian release.
--			Changed over to using Blizzard-defined strings for as many things as possible.
--			Corrected a problem that was causing the tooltip for creatures that needed to be killed to start a quest not to appear properly.
--			Added a few basic localizations.
--			Made the breadcrumb frame hide by default to attempt to eliminate an edge case.
--			Fixed a problem where button clicking behavior would never be set if the button was first entered while in combat.
--			Made prerequisite information appear as question marks instead of causing a LUA error in case the Grail data is lacking.
--		010	Made it so the color of the breadcrumb quest names match their current status.
--			The click areas to the right and bottom of the quest log window no longer extend past the window.
--			Added menu options for Class quests, Profession quests, Reputation quests, and Daily quests.  The Class and Profession quests will show all the quests in the system except for the class and professions that match the player.  For those, the quests are displayed using the normal filtering rules.  The Reputation quests follow the normal filtering rules except those that fail to be acceptable solely because of reputation will be displayed instead of following the display unobtainable filter.
--			Changed over to using Grail's StatusCode() vice Status(), and making use of more modern bit values, thereby requiring version 20.
--			Removed a few event types that are handled because Grail now does that.  Instead switched to using Grail's new Status notification.
--			The tooltips for quests in the panel show profession and reputation requirements as appropriate.
--			Corrected a problem where the quest panel may not update properly while in combat.
--
--	Known Issues
--			Should start displaying in the tooltip the Near and Self NPC information. [At the moment Grail does not support this conveniently.]
--			Attempt to design a system where the prerequisite information is displayed in a cleaner, more accurate manner.
--			For breadcrumbs we may need to switch from the use of GetQuestID() to GetTitleText() (and perhaps need the NPC info) because GetQuestID() may not work as we hope.
--
--	UTF-8 file
--

local Wholly_File_Version = 010
local requiredGrailVersion = 20

--	Set up the bindings to use the localized name Blizzard supplies.  Note that the Bindings.xml file cannot
--	just contain the TOGGLEQUESTLOG because then the entry for Wholly does not show up.  So, we use a version
--	named WHOLLY_TOGGLEQUESTLOG which maps to the same Global string, which works exactly as we want.
BINDING_NAME_WHOLLY_TOGGLEQUESTLOG = BINDING_NAME_TOGGLEQUESTLOG
BINDING_HEADER_WHOLLY = "Wholly"

if nil == Wholly or Wholly.versionNumber < Wholly_File_Version then

	WhollyDatabase = {}

	Wholly = {

		cachedPanelQuests = {},		-- quests and their status for map area currentPanelMapID
		cachedPinQuests = {},		-- quests and their status for map area currentPinMapID
		color = {
			['B'] = "FF996600",	-- brown	[unobtainable]
			['C'] = "FF00FF00",	-- green	[completed]
			['D'] = "FF0099CC",	-- daily	[repeatable]
			['G'] = "FFFFFF00",	-- yellow	[can accept]
			['I'] = "FFFF00FF",	-- purple	[in log]
			['L'] = "FFFFFFFF",	-- white	[too high level]
			['P'] = "FFFF0000",	-- red		[does not meet prerequisites]
			['R'] = "FF0099CC",	-- daily	[true repeatable - used for question mark in pins]
			['U'] = "FF00FFFF",	-- bogus default[unknown]
			['W'] = "FF666666",	-- grey		[low-level can accept]
			},
		configuration = {
			[1] = {[0]="Prepend quest level", [1]='prependsQuestLevel' },
			[2] = {[0]="Append required level", [1]='appendRequiredLevel' },
			[3] = {[0]="Show |cFF0099CCrepeatable quests|r", [1]='showsDailyQuests' },
			[4] = {[0]="Show |cFFFF00FFquests in log|r", [1]='showsQuestsInLog' },
			[5] = {[0]="Show |cFF00FF00completed quests|r", [1]='showsCompletedQuests' },
			[6] = {[0]="Show |cFF996600unobtainable quests|r", [1]='showsUnobtainableQuests' },
			[7] = {[0]="Show |cFF666666low-level quests|r", [1]='showsLowLevelQuests' },
			[8] = {[0]="Show whether repeatable quests previously completed", [1]='showsAnyPreviousRepeatableCompletions' },
			[9] = {[0]="Show high-level quests", [1]='showsHighLevelQuests'},
			[10]= {[0]="Show |cFFFF0000quests that need prerequisites|r", [1]='showsQuestsThatFailPrerequsites'},
			[11]= {[0]="Quest log panel updates when zones change", [1]='updatesPanelWhenZoneChanges'},
			[12]= {[0]="Display map pins for quest givers", [1]='displaysMapPins'},
			[13]= {[0]="Display checkbox on world map", [1]='displaysMapFrame'},
			[14]= {[0]="Display dungeon quests in outer map", [1]='displaysDungeonQuests'},
			[15]= {[0]="Show prerequisite quests in tooltip", [1]='displaysPrerequisites'},
			[16]= {[0]="Display breadcrumb quest information on Quest Frame", [1]='displaysBreadcrumbs'},
			[17]= {[0]="Display holiday quests always", [1]='displaysHolidaysAlways'},
			[18]= {[0]="Open world map updates when zones change", [1]='updatesWorldMapOnZoneChange'},
			[19]= {[0]="Show status of quests in log", [1]='showsInLogQuestStatus'},
			},
		currentPanelMapID = nil,	-- the map area the panel is showing
		currentPinMapID = nil,		-- the map area the world map is showing
		currentPinMapLevel = 0,
		debug = true,
		dropdown = nil,
		dungeonMenuTitles = {},
		eventDispatch = {
			['PLAYER_LOGIN'] = function(self, frame)
				frame:RegisterEvent("WORLD_MAP_UPDATE")			-- this is for pins
				frame:RegisterEvent("ZONE_CHANGED_NEW_AREA")		-- this is for the panel
				self:UpdateBreadcrumb()							-- sets up registration of events for breadcrumbs based on user preferences
			end,
			['PLAYER_REGEN_ENABLED'] = function(self, frame)
				self:ScrollFrame_Update()
				frame:UnregisterEvent("PLAYER_REGEN_ENABLED")
			end,
			['QUEST_DETAIL'] = function(self, frame)
				self:ShowBreadcrumbInfo()
			end,
			['VARIABLES_LOADED'] = function(self, frame)
				if nil == WhollyDatabase.defaultsLoaded then
					self:LoadDefaults()
				end
				if nil == WhollyDatabase.currentSortingMode then
					WhollyDatabase.currentSortingMode = 1
				end

				-- Now to be nicer to those that have used the addon before the current
				-- incarnation, newly added defaults will have their normal setting set
				-- as appropriate.
				if nil == WhollyDatabase.version then		-- first loaded prior to version 006, so default options added in 006
					WhollyDatabase.displaysHolidaysAlways = true		-- version 006
					WhollyDatabase.updatesWorldMapOnZoneChange = true	-- version 006
				elseif WhollyDatabase.version < 7 then
					WhollyDatabase.showsInLogQuestStatus = true			-- version 007
				end
				WhollyDatabase.version = Wholly.versionNumber

				self.mapFrame.check:SetChecked(WhollyDatabase.displaysMapPins)
				self:DisplayMapFrame(WhollyDatabase.displaysMapFrame)
				frame:RegisterEvent("PLAYER_LOGIN")
				Grail:RegisterObserver("Status", self.CallbackHandler)
				Grail:RegisterObserverQuestAbandon(self.CallbackHandler)

				-- Find out which "map area" is for the player's class
				for key, value in pairs(Grail.classMapping) do
					if Grail.playerClass == value then
						self.playerClassMap = Grail.classToMapAreaMapping['C'..key]
					end
				end
			end,
			['WORLD_MAP_UPDATE'] = function(self, frame)
				self:UpdateQuestCaches(false, true)	-- note that we do not force the update since this event happens way too many times
			end,
			['ZONE_CHANGED_NEW_AREA'] = function(self, frame)
				if WhollyDatabase.updatesWorldMapOnZoneChange and WorldMapFrame:IsVisible() then
					local zoneNameToUse = GetRealZoneText()
					if nil ~= zoneNameToUse then
						local newMapId = Grail.zoneNameMapping[zoneNameToUse]
						if nil ~= newMapId then
							SetMapByID(newMapId)
						end
					end
				end
				self:UpdateQuestCaches(false, false, WhollyDatabase.updatesPanelWhenZoneChanges)

				-- Now update open tooltips showing our quest count data
				if GameTooltip:IsVisible() and GameTooltip:GetOwner() == com_mithrandir_whollyFrameSwitchZoneButton then
					GameTooltip:ClearLines()
					GameTooltip:AddLine(Wholly.panelCountLine)
				elseif self.tooltip:IsVisible() and self.tooltip:GetOwner() == com_mithrandir_WhollyCheckButton then
					self.tooltip:ClearLines()
					self.tooltip:AddLine(Wholly.mapCountLine)
				elseif GameTooltip:IsVisible() and GameTooltip:GetOwner() == self.ldbTooltipOwner then -- LibDataBroker tooltip
					GameTooltip:ClearLines()
					GameTooltip:AddLine("Wholly - " .. UIDropDownMenu_GetText(Wholly.dropdown) )
					GameTooltip:AddLine(Wholly.panelCountLine)					
				end
			end,
			},
		filteredPanelQuests = {},	-- filtered table from cachedPanelQuests using current panel filters
		filteredPinQuests = {},		-- filtered table from cachedPinQuests using current pin filters
		initialUpdateProcessed = false,
		lastPrerequisiteQuest = nil,
		ldbTooltipOwner = nil,
		mapFrame = nil,			-- the world map frame that contains the checkbox to toggle pins
		npcs = {},
		pairedConfigurationButton = nil,-- configuration panel button that does the same thing as the world map button
		playerClassMap = nil,
		regFrames = {},
		regNPCPins = {},		-- in unifying the pins we keep a table of pins associated with NPC id
		regPins = {},
		reputationLevelMapping = { [0] = 1, [35999] = 2, [38999] = 3, [41999] = 4, [44999] = 5, [50999] = 6, [62999] = 7, [83999] = 8, [84998] = 8, },
		s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Turn in:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			},
		timeSinceLastUpdate = 0,
		tooltip = nil,
		updateDelay = 0.5,
		versionNumber = Wholly_File_Version,

		AddPrerequisite = function(self, indentation, preqs)
			local shown = false
			for i = 1, #preqs do
				if shown then GameTooltip:AddLine(" ") end
				local questId = preqs[i][1][1]
				local status = self:ClassificationOfQuest(questId)
				if 'B' == status or 'C' == status then return end
				local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questId, 'A', nil, nil, true)
				local firstInfo = self:PrettyQuestString({ questId, status })
				local secondInfo = ""
				lastPrerequisiteQuest = questId
				if 'I' == status then lastPrerequisiteQuest = nil end
				if nil ~= locations then
					local printed = false
					for i = 1, #locations do
						local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
						if xcoord ~= nil then
							npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
							secondInfo = npcname .. " " .. Grail.mapAreaMapping[mapId] .. " " .. xcoord .. " " .. ycoord
							if printed then firstInfo = " " end
							GameTooltip:AddDoubleLine(firstInfo, secondInfo)
							printed = true
						end
					end
					if not printed then
						GameTooltip:AddLine(firstInfo)
					end
				else
					GameTooltip:AddLine(firstInfo)
				end
				shown = true
				if nil ~= preqs[i][1][2] then
					self:AddPrerequisite(indentation + 1, preqs[i][1][2])
				end
			end
		end,

		BreadcrumbClick = function(self, frame)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			local questId = GetQuestID()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			if nil ~= breadcrumbs then
				for j = 1, #breadcrumbs do
					local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(breadcrumbs[j], 'A', nil, nil, true)
					if nil ~= locations then
						for i = 1, #locations do
							local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], 	locations[i][6], locations[i][7]
							if xcoord ~= nil then
								npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
								TomTom:AddMFWaypoint(mapId, dungeonLevel, xcoord/100, ycoord/100, { title = npcname })
							end
						end
					end
				end
			end
		end,

		BreadcrumbEnter = function(self, frame)
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:ClearLines()
			local questId = GetQuestID()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			if nil ~= breadcrumbs then
				GameTooltip:AddLine(self.s['BREADCRUMB'])
				for i = 1, #breadcrumbs do
					GameTooltip:AddLine(self:PrettyQuestString({ breadcrumbs[i], self:ClassificationOfQuest(breadcrumbs[i]) }))
				end
				GameTooltip:Show()
			end
		end,

		ButtonEnter = function(self, button, ...)
			GameTooltip:SetOwner(button, "ANCHOR_RIGHT")
			GameTooltip:ClearLines()
			GameTooltip:SetHyperlink(format("quest:%d", button.questId))

			if not button.secureProcessed and not InCombatLockdown() then
				button:SetAttribute("type1", "click")
				button:SetAttribute("clickbutton", Wholly)
				button:SetAttribute("type2", "click")
				button:SetAttribute("alt-type1", "macro")
				button.secureProcessed = true
			else
				-- TODO: Should attempt a delayed setting of this if not button.secureProcessed and InCombatLockdown()
			end

			if self.debug then
				GameTooltip:AddLine(" ")
				GameTooltip:AddDoubleLine("Quest ID:"..button.questId, button.reason)
			end

			local codes = Grail:CodesWithPrefixQuest(button.questId, 'K')
			if nil ~= codes and strlen(codes[1]) > 4 then
				local outputString = ""
				local bitValue = tonumber(strsub(codes[1], 5))
				if bitValue > 0 then
					if bit.band(bitValue, 1) > 0 then outputString = outputString .. self.s['REPEATABLE'] .. " " end
					if bit.band(bitValue, 2) > 0 then outputString = outputString .. DAILY .. " " end
					if bit.band(bitValue, 4) > 0 then outputString = outputString .. CALENDAR_REPEAT_WEEKLY .. " " end
					if bit.band(bitValue, 8) > 0 then outputString = outputString .. CALENDAR_REPEAT_MONTHLY .. " " end
					if bit.band(bitValue, 16) > 0 then outputString = outputString .. self.s['YEARLY'] .. " " end
					if bit.band(bitValue, 32) > 0 then outputString = outputString .. self.s['ESCORT'] .. " " end
					if bit.band(bitValue, 64) > 0 then outputString = outputString .. CALENDAR_TYPE_DUNGEON .. " " end
					if bit.band(bitValue, 128) > 0 then outputString = outputString .. CALENDAR_TYPE_RAID .. " " end
					if bit.band(bitValue, 256) > 0 then outputString = outputString .. CALENDAR_TYPE_PVP .. " " end
					if bit.band(bitValue, 512) > 0 then outputString = outputString .. CHANNEL_CATEGORY_GROUP .. " " end
					if bit.band(bitValue, 1024) > 0 then outputString = outputString .. PLAYER_DIFFICULTY2 .. " " end
					GameTooltip:AddLine(" ")
					GameTooltip:AddLine(outputString)
				end
			end

			codes = Grail:CodesWithPrefixQuest(button.questId, 'P')
			if nil ~= codes then
				for i = 1, #codes do
					local controlValue = strsub(codes[i], 2, 2)
					local colonCheck = strsub(codes[i], 3, 3)
					if ':' == controlValue or ('L' == controlValue and ':' == colonCheck) or ('H' == controlValue and ':' == colonCheck)  or ('C' == controlValue and ':' == colonCheck) then
						-- ignore these since they are not profession codes
					else
						local profValue = tonumber(strsub(codes[i], 3, 5))
						GameTooltip:AddDoubleLine(self.s['PROFESSION_REQUIRED'], "|cFF00FF00"..Grail.professionMapping[controlValue].." "..profValue.."|r")
					end
				end
			end

			local factionStandingFormat = "FACTION_STANDING_LABEL%d"
			if Grail.playerGender == 3 then factionStandingFormat = factionStandingFormat.."_FEMALE" end
			-- Now reputation requirements
			codes = Grail:CodesWithPrefixQuest(button.questId, 'V')
			if nil ~= codes then
				for i = 1, #codes do
					local repIndex = tonumber(strsub(codes[i], 2, 3))
					local repValue = tonumber(strsub(codes[i], 4, 8))
					GameTooltip:AddDoubleLine(self.s['REPUTATION_REQUIRED'], format("|cFF00FF00%s %s|r", Grail.reputationMapping[repIndex], GetText(format(factionStandingFormat, self.reputationLevelMapping[repValue]))))
				end
			end
			codes = Grail:CodesWithPrefixQuest(button.questId, 'W')
			if nil ~= codes then
				for i = 1, #codes do
					local repIndex = tonumber(strsub(codes[i], 2, 3))
					local repValue = tonumber(strsub(codes[i], 4, 8))
					GameTooltip:AddDoubleLine(self.s['REPUTATION_REQUIRED'], "|cFFFF0000"..Grail.reputationMapping[repIndex].." "..repValue.."|r")
				end
			end

			local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(button.questId, 'A', nil, nil, true)
			if nil ~= locations then
				GameTooltip:AddLine(" ")
				GameTooltip:AddLine(self.s['QUESTGIVERS'])
				for i = 1, #locations do
					local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
					if xcoord ~= nil then
						if 'I' ~= button.statusCode then
							local desiredMacroValue = SLASH_TARGET1 .. " " .. npcname
							if button:GetAttribute("macrotext") ~= desiredMacroValue then
								if not InCombatLockdown() then
									button:SetAttribute("macrotext", desiredMacroValue)
								else
									-- TODO: Should attempt a delayed setting of this
								end
							end
						end
						npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
						if nil == npcname then
							print("Wholly problem: No NPC name for Giver for Quest", button.questId)
							npcname = '????'
						end
						local mapName = Grail.mapAreaMapping[mapId]
						if nil == mapName then
							print("Wholly problem: No map for Giver for Quest", button.questId)
							mapName = '???'
						end
						if nil == ycoord then
							print("Wholly problem: No ycoord for Giver for Quest", button.questId)
							ycoord = '??'
						end
						GameTooltip:AddDoubleLine(npcname, mapName .. " " .. xcoord .. " " .. ycoord)
					end
				end
			end
			locations, mustKill = Grail:MultipleUniqueMapLocationQuest(button.questId, 'T', nil, nil, true)
			if nil ~= locations then
				GameTooltip:AddLine(" ")
				GameTooltip:AddLine(self.s['TURNINS'])
				for i = 1, #locations do
					local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
					if xcoord ~= nil then
						if 'I' == button.statusCode then
							local desiredMacroValue = SLASH_TARGET1 .. " " .. npcname
							if button:GetAttribute("macrotext") ~= desiredMacroValue then
								if not InCombatLockdown() then
									button:SetAttribute("macrotext", desiredMacroValue)
								else
									-- TODO: Should attempt a delayed setting of this
								end
							end
						end
						npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
						if nil == npcname then
							print("Wholly problem: No NPC name for turn in for Quest", button.questId)
							npcname = '????'
						end
						local mapName = Grail.mapAreaMapping[mapId]
						if nil == mapName then
							print("Wholly problem: No map for turn in for Quest", button.questId)
							mapName = '???'
						end
						if nil == ycoord then
							print("Wholly problem: No ycoord for turn in for Quest", button.questId)
							ycoord = '??'
						end
						GameTooltip:AddDoubleLine(npcname, mapName .. " " .. xcoord .. " " .. ycoord)
					end
				end
			end


			local breadcrumbs = Grail:AvailableBreadcrumbs(button.questId)
			if nil ~= breadcrumbs then
				GameTooltip:AddLine(" ")
				GameTooltip:AddLine(self.s['BREADCRUMB'])
				for i = 1, #breadcrumbs do
					GameTooltip:AddLine(self:PrettyQuestString({ breadcrumbs[i], self:ClassificationOfQuest(breadcrumbs[i]) }))
				end
			end

			lastPrerequisiteQuest = nil
			if 'P' == button.statusCode and WhollyDatabase.displaysPrerequisites then	-- prerequisites
				local _, preqs = Grail:AncestorQuests(button.questId)
				if nil ~= preqs and 0 < #preqs then
					GameTooltip:AddLine(" ")
					GameTooltip:AddLine(self.s['PREREQUISITES'])
					self:AddPrerequisite(0, preqs)
				end
			end

			GameTooltip:Show()
		end,

		ButtonPostClick = function(self, button)
			if button ~= self.clickingButton then print("Post click not from the same Pre click") end
			self.clickingButton = nil
		end,

		ButtonPreClick = function(self, button)
			self.clickingButton = button
		end,

		print_r = function(self, t, indent, done)
			done = done or {}
			indent = indent or ''
			local nextIndent
			for key, value in pairs (t) do
				if type (value) == "table" and not done[value] then
					nextIndent = nextIndent or (indent .. string.rep(' ',string.len(tostring (key))+2))
					done [value] = true
					print(indent .. "[" .. tostring (key) .. "] => Table {");
					print(nextIndent .. "{");
					self:print_r(value, nextIndent .. string.rep(' ',2), done)
					print(nextIndent .. "}")
				else
					print(indent .. "[" .. tostring (key) .. "] => " .. tostring (value).."")
				end
			end
		end,

		CallbackHandler = function(type, questId)
			Wholly:UpdateQuestCaches(true)
		end,

		CheckNPCTooltip = function(tooltip)
			if (not UnitIsPlayer("mouseover") or true) then
				-- check if this npc drops a quest item
				local guid = UnitGUID("mouseover");
				if ( guid ) then
					local targetType = tonumber(guid:sub(5,5), 16) % 8
					local id = tonumber(guid:sub(7, 10), 16);
					if 1 == targetType then id = id + 1000000 end	-- world object
					local qs = Wholly.npcs[id]
					if nil ~= qs then
						for j = 1, #(qs) do
							local q = qs[j]
							if q then
								local status = Grail:StatusCode(q)
								if 0 == bit.band(status, Grail.bitMaskAcceptableMask - Grail.bitMaskPrerequisites - Grail.bitMaskLevelTooLow) then
									local _, kindsOfNPC = Grail:IsTooltipNPC(id)
									if nil ~= kindsOfNPC then
										for i = 1, #(kindsOfNPC), 1 do
											local tooltipMessage = nil
											if kindsOfNPC[i][1] == Grail.NPC_TYPE_KILL then
												tooltipMessage = format(Wholly.s['KILL_TO_START_FORMAT'], Grail:QuestName(q))
											elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
												if Wholly:DroppedItemMatchesQuest(kindsOfNPC[i][2], q) then
													tooltipMessage = format(Wholly.s['DROP_TO_START_FORMAT'], Grail:NPCName(kindsOfNPC[i][2]), Grail:QuestName(q))
												end
											end
											if nil ~= tooltipMessage then
												local leftStr = format("|TInterface\\MINIMAP\\ObjectIcons:17:9:0:0:256:128:40:56:32:63|t %s", tooltipMessage)
												tooltip:AddLine(leftStr);
											end
										end
									end
									tooltip:Show();
								end
							end
						end
					end
				end
			end
		end,

		ClassificationOfQuest = function(self, questId, shouldDisplayHolidays)
			local retval = 'U'
			local questBitMask
			if nil ~= questId and tonumber(questId) then
				questId = tonumber(questId)
				questBitMask = Grail:StatusCode(questId)
				if bit.band(questBitMask, Grail.bitMaskNonexistent + Grail.bitMaskError) > 0 then
					retval = 'U'
				elseif bit.band(questBitMask, Grail.bitMaskQuestFailureWithAncestor - (shouldDisplayHolidays and Grail.bitMaskHoliday or 0) ) > 0 then
					retval = 'B'
				elseif bit.band(questBitMask, Grail.bitMaskCompleted) > 0 then
					retval = 'C'
				elseif bit.band(questBitMask, Grail.bitMaskInLog) > 0 then
					retval = 'I'
				elseif bit.band(questBitMask, Grail.bitMaskPrerequisites) > 0 then
					retval = 'P'
				elseif bit.band(questBitMask, Grail.bitMaskResettable) > 0 then
					retval = 'D'
				elseif bit.band(questBitMask, Grail.bitMaskRepeatable) > 0 then
					retval = 'R'
				elseif bit.band(questBitMask, Grail.bitMaskLevelTooLow) > 0 then
					retval = 'L'
				elseif bit.band(questBitMask, Grail.bitMaskLowLevel) > 0 then
					retval = 'W'
				else
					retval = 'G'
				end
			end
			if self.debug then
				return retval, self:QuestFailureString(questBitMask)
			else
				return retval
			end
		end,

		---
		--	Gets all the quests in the map area, then classifies them based on the current player.
		ClassifyQuestsInMap = function(self, mapId)
			local retval = nil
			if nil ~= mapId and tonumber(mapId) then
				mapId = tonumber(mapId)
				local displaysHolidayQuestsAlways = false
				if mapId > 100000 and mapId < 200000 and WhollyDatabase.displaysHolidaysAlways then displaysHolidayQuestsAlways = true end
				retval = {}
				local questsInMap = Grail:QuestsInMap(mapId, false, false, WhollyDatabase.displaysDungeonQuests) or {}
				for _,questId in pairs(questsInMap) do
					tinsert(retval, { questId, self:ClassificationOfQuest(questId, displaysHolidayQuestsAlways) })
				end
			end
			return retval
		end,

		-- This is named this way with this function signature because it is called from the SecureActionButtonTemplate exactly like this.
		Click = function(self, leftOrRight)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			local button = self.clickingButton
			local questIdToUse = button.questId
			local npcType = 'A'
			if "RightButton" == leftOrRight then
				if nil ~= lastPrerequisiteQuest then
					questIdToUse = lastPrerequisiteQuest
				elseif 'I' == button.statusCode then
					npcType = 'T'
				end
			end
			local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questIdToUse, npcType, nil, nil, true)
			if nil ~= locations then
				for i = 1, #locations do
					local xcoord, ycoord, npcname, npcId, mapId, dungeonLevel, realAreaId = locations[i][1], locations[i][2], locations[i][3], locations[i][4], locations[i][5], locations[i][6], locations[i][7]
					if xcoord ~= nil then
						npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
						TomTom:AddMFWaypoint(mapId, dungeonLevel, xcoord/100, ycoord/100, { title = npcname })
					end
				end
			end
		end,

		ConfigFrame_OnLoad = function(self, panel)
			panel.name = "Wholly"
			panel:Hide()
			InterfaceOptions_AddCategory(panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				local button = CreateFrame("CheckButton", parent.."Button"..i, panel, "InterfaceOptionsCheckButtonTemplate")
				button:SetPoint("TOPLEFT", panel, "TOPLEFT", 8, (i * -20) + 10)
				_G[button:GetName().."Text"]:SetText(self.configuration[i][0])
				button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:ScrollFrame_Update() Wholly:UpdatePins() end)

				-- Special case for the button that matches the same functionality as the one on the world map
				if 12 == i then
					self.pairedConfigurationButton = button
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly.mapFrame.check:SetChecked(self:GetChecked()) Wholly:UpdatePins() end)
				end

				-- Special case for the button that controls the display of the world map frame
				if 13 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:DisplayMapFrame(self:GetChecked()) end)
				end

				-- Special case for the button that controls whether quests in dungeons appear in the outer map
				-- This is done because the initial load of quests in an area which would normally get filtered by the default behavior needs to be reloaded
				if 14 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:UpdateQuestCaches(true) Wholly:ScrollFrame_Update() Wholly:UpdatePins() end)
				end

				-- Special case for the breadcrumb control button
				if 16 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly:UpdateBreadcrumb(Wholly) end)
				end

				-- Special case the holiday button
				if 17 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked() Wholly.ForcePanelMapArea(Wholly) Wholly:ScrollFrame_Update() end)
				end

				-- Special case the button that shows in log quest status
				if 19 == i then
					button:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked()
									Wholly:ScrollFrame_Update()
									end)
				end

			end
		end,

		ConfigFrame_OnShow = function(self, panel)
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				_G[parent.."Button"..i]:SetChecked(WhollyDatabase[self.configuration[i][1]])
			end
		end,

		DisplayMapFrame = function(self, shouldDisplay)
			if shouldDisplay then self.mapFrame:Show() else self.mapFrame:Hide() end
		end,

		Distance = function(self, x1, y1, x2, y2)
			-- gets distance in map units TODO: use yards?
			local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();

			return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
		end,

		Dropdown_AddButton = function(buttonTitle, level)
			local info = UIDropDownMenu_CreateInfo()
			info.hasArrow = true
			info.notCheckable = true
			info.text = buttonTitle
			info.value = { ["Level1_Key"] = buttonTitle }
			UIDropDownMenu_AddButton(info, level)
		end,

		---
		--	returns a table that is structured like the continents, sorted in localized name order
		Dropdown_Array = function(self, mapIdTable)
			local retval = {}
			for i = 1, #mapIdTable do
				tinsert(retval, { ["name"] = Grail.mapAreaMapping[mapIdTable[i]], ["mapID"] = mapIdTable[i] })
			end
			table.sort(retval, function(a, b) return a["name"] < b["name"] end)
			return retval
		end,

		Dropdown_Initialize = function(self, level)
			level = level or 1
			local info
			local zoneCount
			local indexToUse
			if 1 == level then
				for i = 1, #(Grail.continents) do
					Wholly.Dropdown_AddButton(Grail.continents[i]["name"], level)
					zoneCount = #(Grail.continents[i]["zones"]) 
					if 4 == zoneCount then indexToUse = 5
					elseif 28 == zoneCount then indexToUse = 1
					elseif 39 == zoneCount then indexToUse = 2
					elseif 8 == zoneCount then indexToUse = 3
					elseif 12 == zoneCount then indexToUse = 4
					end
					Wholly.dungeonMenuTitles[indexToUse] = BUG_CATEGORY3 .. ", " .. Grail.continents[i]["name"]
				end

				-- Dungeons
				Wholly.Dropdown_AddButton(Wholly.dungeonMenuTitles[1], level)
				Wholly.Dropdown_AddButton(Wholly.dungeonMenuTitles[2], level)
				Wholly.Dropdown_AddButton(Wholly.dungeonMenuTitles[3], level)
				Wholly.Dropdown_AddButton(Wholly.dungeonMenuTitles[4], level)
				Wholly.Dropdown_AddButton(Wholly.dungeonMenuTitles[5], level)

				-- World Events
				Wholly.Dropdown_AddButton(Wholly.s['WORLD_EVENTS'], level)

				-- Class
				Wholly.Dropdown_AddButton(CLASS, level)

				-- Professions
				Wholly.Dropdown_AddButton(TRADE_SKILLS, level)

				-- Reputation
				Wholly.Dropdown_AddButton(REPUTATION, level)

				-- Daily
				info = UIDropDownMenu_CreateInfo()
				info.hasArrow = false
				info.notCheckable = true
				info.text = DAILY
				info.value = { ["mapID"] = 400000 }
				info.func = function()
					Wholly.currentPanelMapID = 400000
					Wholly.ForcePanelMapArea(Wholly)
					CloseDropDownMenus()
				end
				UIDropDownMenu_AddButton(info, level)

				-- Other
				Wholly.Dropdown_AddButton(Wholly.s['OTHER'], level)

			elseif 2 == level then
				local Level1_Key = UIDROPDOWNMENU_MENU_VALUE["Level1_Key"]
				local zones = {}
				for i = 1, #(Grail.continents) do
					if Level1_Key == Grail.continents[i]["name"] then
						zones = Grail.continents[i]["zones"]
					end
				end
				if Level1_Key == Wholly.dungeonMenuTitles[1] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonKalimdorMapping)
				end
				if Level1_Key == Wholly.dungeonMenuTitles[2] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonEasternKingdomsMapping)
				end
				if Level1_Key == Wholly.dungeonMenuTitles[3] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonOutlandMapping)
				end
				if Level1_Key == Wholly.dungeonMenuTitles[4] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonNorthrendMapping)
				end
				if Level1_Key == Wholly.dungeonMenuTitles[5] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.dungeonMaelstromMapping)
				end
				if Level1_Key == Wholly.s['WORLD_EVENTS'] then
					for code, name in pairs(Grail.holidayMapping) do
						tinsert(zones, { ["name"] = name, ["mapID"] = Grail.holidayToMapAreaMapping['H'..code] })
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == CLASS then
					for code, englishName in pairs(Grail.classMapping) do
						local localizedGenderClassName = Grail:CreateClassNameLocalizedGenderized(englishName)
						local classColor = RAID_CLASS_COLORS[englishName]
						tinsert(zones, { ["name"] = format("|cff%.2x%.2x%.2x%s|r", classColor.r*255, classColor.g*255, classColor.b*255, localizedGenderClassName), ["mapID"] = Grail.classToMapAreaMapping['C'..code], ["nonColorName"] = localizedGenderClassName })
					end
					table.sort(zones, function(a, b) return a["nonColorName"] < b["nonColorName"] end)
				end
				if Level1_Key == TRADE_SKILLS then
					for code, professionName in pairs(Grail.professionMapping) do
						local mapId = Grail.professionToMapAreaMapping['P'..code]
						if nil ~= Grail.mapAreaMapping[mapId] then
							tinsert(zones, { ["name"] = professionName, ["mapID"] = mapId })
						end
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == REPUTATION then
					for reputationIndex, reputationName in pairs(Grail.reputationMapping) do
						local mapId = 400000 + reputationIndex
						if nil ~= Grail.mapAreaMapping[mapId] then
							tinsert(zones, { ["name"] = reputationName, ["mapID"] = mapId })
						end
					end
					table.sort(zones, function(a, b) return a["name"] < b["name"] end)
				end
				if Level1_Key == Wholly.s['OTHER'] then
					zones = Wholly.Dropdown_Array(Wholly, Grail.otherMapping)
				end
				for i = 1, #zones do
					info = UIDropDownMenu_CreateInfo()
					info.hasArrow = false
					info.notCheckable = true
					info.text = zones[i]["name"]
					info.value = { ["mapID"] = zones[i]["mapID"] }
					info.func = function()
						Wholly.currentPanelMapID = zones[i]["mapID"]
						Wholly.ForcePanelMapArea(Wholly)
						CloseDropDownMenus()
					end
					UIDropDownMenu_AddButton(info, level)
				end
			end
		end,

		DroppedItemMatchesQuest = function(self, dropNPCCode, matchingQuestId)
			local retval = true
			local questCodes = Grail:CodesWithPrefixNPC(dropNPCCode, "Q:")
			if nil ~= questCodes then
				retval = false
				for i = 1, #questCodes do
					local quests = { strsplit(",", strsub(questCodes[i], 3)) }
					for j = 1, #quests do
						if tonumber(quests[j]) == tonumber(matchingQuestId) then retval = true end
					end
				end
			end
			return retval
		end,

		FilterQuests = function(self, forPanel)
			local f = forPanel and self.filteredPanelQuests or self.filteredPinQuests
			f = {}
			local questsInMap = forPanel and self.cachedPanelQuests or self.cachedPinQuests
			local shouldAdd, statusCode

			--	We want to be able to force display of quests that are class or profession specific
			--	unless they are associated with the player.  In that case, the display of the quests
			--	obeys the same rules as the quests in a normal map area.
			local shouldForce = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID > 200000 and self.currentPanelMapID < 400000 then
				shouldForce = true
				if self.playerClassMap == self.currentPanelMapID then shouldForce = false end
				if self.currentPanelMapID > 300000 then
					for key,value in pairs(Grail.professionToMapAreaMapping) do
						if value == self.currentPanelMapID then
							local actualKey = key:sub(2, 2)
							if Grail:ProfessionExceeds(actualKey, 1) then -- indicates the profession is known
								shouldForce = false
							end
						end
					end
				end
			end
			local repuationQuest = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID > 400000 and self.currentPanelMapID < 500000 then
				reputationQuest = true
			end

			for i = 1, #questsInMap do
				statusCode = questsInMap[i][2]

				if 'W' == statusCode then shouldAdd = WhollyDatabase.showsLowLevelQuests
				elseif 'I' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsQuestsInLog
				elseif 'C' == statusCode then shouldAdd = forPanel and WhollyDatabase.showsCompletedQuests
				elseif 'B' == statusCode then
					local reputationForce = false
					if reputationQuest then
						local status = Grail:StatusCode(questsInMap[i][1])
						local bitResult = bit.band(status, Grail.bitMaskQuestFailureWithAncestor - (Grail.bitMaskAncestorReputation + Grail.bitMaskReputation))
						if bitResult == 0 then		-- the quest only failed because of reputation failure
							reputationForce = true
						end
					end
					shouldAdd = forPanel and (WhollyDatabase.showsUnobtainableQuests or shouldForce or reputationForce)
				elseif 'D' == statusCode then shouldAdd = WhollyDatabase.showsDailyQuests
				elseif 'L' == statusCode then shouldAdd = WhollyDatabase.showsHighLevelQuests
				elseif 'P' == statusCode then shouldAdd = WhollyDatabase.showsQuestsThatFailPrerequsites
				else shouldAdd = true
				end

				if shouldAdd then
					tinsert(f, questsInMap[i])
				end
			end
			if forPanel then self.filteredPanelQuests = f else self.filteredPinQuests = f end
			if not forPanel then
				self.mapCountLine = self:PrettyQuestCountString(questsInMap)
			else
				self.panelCountLine = self:PrettyQuestCountString(questsInMap)
			end
		end,

		FilterPanelQuests = function(self)
			self:FilterQuests(true)
		end,

		FilterPinQuests = function(self)
			self:FilterQuests(false)
		end,

		ForcePanelMapArea = function(self)
			local mapAreaName = Grail.mapAreaMapping[self.currentPanelMapID] or GetRealZoneText()	-- default to something if we do not support the zone
			if nil ~= mapAreaName then UIDropDownMenu_SetText(self.dropdown, mapAreaName) end
			self.cachedPanelQuests = self:ClassifyQuestsInMap(self.currentPanelMapID) or {}
			if not InCombatLockdown() then
				self:ScrollFrame_Update()
			else
				com_mithrandir_whollyFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			end
		end,

		GetMousePosition = function(self)
			local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
			local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
			local scale = WorldMapDetailFrame:GetEffectiveScale();

			local x, y = GetCursorPosition();
			local cx = (x/scale - left) / width;
			local cy = (top - y/scale) / height;
	
			return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
		end,

		GetPin = function(self, npcId)
			if nil ~= self.regNPCPins[npcId] then return self.regNPCPins[npcId] end		-- already have a pin for this npcId

			if #self.regFrames > 0 then
				local pin = table.remove(self.regFrames)
				pin.npcId = npcId
			end

			local pin = CreateFrame("Frame", nil, WorldMapDetailFrame);
			pin.npcId = npcId
			pin:SetWidth(16);
			pin:SetHeight(16);
			pin:EnableMouse(true);
			pin:SetScript("OnEnter", function(pin) self:ShowTooltip(pin) end)
			pin:SetScript("OnLeave", function() self:HideTooltip() end)
			pin.SetType = function(self, texType)
				if self.texType == texType then return end -- don't need to make changes

				-- 256x64, 8x4
				self.texture:SetDesaturated(false);
				self.texture:SetVertexColor(1, 1, 1);
				if texType == "D" then
					self.texture:SetTexCoord(3*0.125, 4*0.125, 1*0.250, 2*0.250);
				elseif texType == "R" then
					self.texture:SetTexCoord(4*0.125, 5*0.125, 1*0.250, 2*0.250);
				elseif texType == "P" then
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetVertexColor(1.0, 0.0, 0.0);
				elseif texType == "W" then
					-- I don't like this one, it's too hard to see.
					-- self.texture:SetTexCoord(4*0.125, 5*0.125, 3*0.250, 4*0.250); -- this is the actual dimmed quest giver icon
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetVertexColor(0.75, 0.75, 0.75);
				elseif texType == "L" then
					-- Note: not all systems can desaturate. If they don't, I don't really care.
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
					self.texture:SetDesaturated(1);
				else
					self.texture:SetTexCoord(1*0.125, 2*0.125, 1*0.250, 2*0.250);
				end

				self.texType = texType
			end

			pin.texture = pin:CreateTexture();
			pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
			pin.texture:SetAllPoints();
			pin:SetType("L")
			self.regNPCPins[npcId] = pin
			return pin;
		end,

		HideAllPins = function(self)
			for i, v in pairs(self.regPins) do
				self:HidePin(i)
			end
		end,

		HidePin = function(self, id)
			if self.regPins[id] then
				self.regPins[id]:Hide()
				table.insert(self.regFrames, self.regPins[id])
				self.regNPCPins[self.regPins[id].npcId] = nil
				self.regPins[id] = nil
			end
		end,

		HideTooltip = function(self)
			self.tooltip:Hide()
		end,

		LoadDefaults = function(self)
			local db = {}
			db.defaultsLoaded = true
			db.prependsQuestLevel = true
			db.appendRequiredLevel = true
			db.showsLowLevelQuests = true
			db.showsAnyPreviousRepeatableCompletions = true
			db.updatesPanelWhenZoneChanges = true
			db.displaysMapPins = true
			db.displaysMapFrame = true
			db.displaysDungeonQuests = true
			db.displaysPrerequisites = true
			db.displaysBreadcrumbs = true
			db.displaysHolidaysAlways = true
			db.updatesWorldMapOnZoneChange = true
			db.showsInLogQuestStatus = true
			db.version = Wholly.versionNumber
			WhollyDatabase = db
		end,

		OnEvent = function(self, frame, event, ...)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, frame, ...)
			end
		end,

		OnHide = function(self, frame)
		end,

		OnLoad = function(self, frame)
			if not Grail or Grail.versionNumber < requiredGrailVersion then print(format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)) return end
			SlashCmdList["WHOLLY"] = function(msg)
				self:SlashCommand(frame, msg)
			end
			SLASH_WHOLLY1 = "/wholly"
			com_mithrandir_whollyFrameTitleText:SetText("Wholly ".. com_mithrandir_whollyFrameTitleText:GetText())

			com_mithrandir_whollyFrameSortButton:SetText(self.s['SORT'])

			-- The frame is not allowing button presses to things just on the outside of its bounds so we move the hit rect
			frame:SetHitRectInsets(0, 32, 0, 84)

			local LibStub = _G["LibStub"]
			if LibStub then
				local LDB = LibStub("LibDataBroker-1.1", true)
				if LDB then
					local launcher = LDB:NewDataObject("Wholly", { type="launcher", icon="Interface\\Icons\\INV_Misc_Book_07",
							OnClick = function(theFrame, button) if button == "RightButton" then InterfaceOptionsFrame_OpenToCategory("Wholly") else frame:Show() end end,
							OnTooltipShow = function(tooltip) Wholly:ProcessInitialUpdate() Wholly.ldbTooltipOwner = tooltip:GetOwner()
							tooltip:AddLine("Wholly - " .. UIDropDownMenu_GetText(Wholly.dropdown) ) tooltip:AddLine(Wholly.panelCountLine) end, 
							})
				end
			end

			self.dropdown = CreateFrame("Button", frame:GetName().."ZoneButton", frame, "UIDropDownMenuTemplate")
			UIDropDownMenu_Initialize(self.dropdown, self.Dropdown_Initialize) -- took away "MENU" because no show with it
			self.dropdown:SetPoint("TOPLEFT", frame, "TOPLEFT", 60, -40)
			UIDropDownMenu_SetWidth(self.dropdown, 240, 0)
			UIDropDownMenu_JustifyText(self.dropdown, "LEFT")
			-- By default, the dropdown has it clicking work with the little button on the right.  This makes it work for the whole button:
			self.dropdown:SetScript("OnClick", function(self) ToggleDropDownMenu(nil, nil, Wholly.dropdown) PlaySound("igMainMenuOptionCheckBoxOn") end)

			self.tooltip = CreateFrame("GameTooltip", "com_mithrandir_WhollyTooltip", UIParent, "GameTooltipTemplate");
			self.tooltip:SetFrameStrata("TOOLTIP");
			self.tooltip.large = com_mithrandir_WhollyTooltipTextLeft1:GetFontObject();
			self.tooltip.small = com_mithrandir_WhollyTooltipTextLeft2:GetFontObject();
			self.tooltip.SetLastFont = function(self, fontObj, rightText)
				local txt = rightText and "Right" or "Left"
				_G[format("com_mithrandir_WhollyTooltipText%s%d", txt, self:NumLines())]:SetFont(fontObj:GetFont())
			end

			-- Create the checkbox in the world map
			local f = CreateFrame("Frame", nil, WorldMapDetailFrame)
			f:SetSize(2, 20)
			f:SetPoint("TOPLEFT", WorldMapDetailFrame, "TOPLEFT", 5, -2)
			f:Hide()
			f:EnableMouse(1)
			f:SetFrameStrata("HIGH")
			f:SetFrameLevel(5)
			f.check = CreateFrame("CheckButton", "com_mithrandir_WhollyCheckButton", f, "OptionsBaseCheckButtonTemplate")
			f.check:SetSize(18, 18)
			f.check:SetHitRectInsets(0, 0, 0, 0)
			f.check:SetPoint("LEFT")
			f.check:SetFrameLevel(6)
			f.check:SetScript("OnClick", function(self) WhollyDatabase[Wholly.configuration[12][1]] = self:GetChecked() Wholly.pairedConfigurationButton:SetChecked(self:GetChecked()) Wholly:UpdatePins() end)
			f.check:SetScript("OnEnter", function(self)
				Wholly.tooltip:ClearLines()
				Wholly.tooltip:SetOwner(self)
				Wholly.tooltip:AddLine(Wholly.mapCountLine)
				Wholly.tooltip:Show()
				Wholly.tooltip:ClearAllPoints()
				Wholly.tooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 20, 0)
				end)
			f.check:SetScript("OnLeave", function(self) Wholly.tooltip:Hide() end)
			local font, size = GameTooltipHeader:GetFont()
			f.text = f:CreateFontString()
			f.text:SetFont(font, size, "OUTLINE")
			f.text:SetTextColor(1, 210/255, 0, 1)
			f.text:SetText("Wholly")
			f.text:SetPoint("LEFT", f.check, "RIGHT")
			self.mapFrame = f

			-- if the UI panel disappears (maximized WorldMapFrame) we need to change parents
			UIParent:HookScript("OnHide", function()
				self.tooltip:SetParent(WorldMapFrame);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)
			UIParent:HookScript("OnShow", function()
				self.tooltip:SetParent(UIParent);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)

			GameTooltip:HookScript("OnTooltipSetUnit", Wholly.CheckNPCTooltip)

			frame:SetScript("OnEvent", function(frame, event, ...) self:OnEvent(frame, event, ...) end)
			frame:RegisterEvent("VARIABLES_LOADED")

		end,

		---
		--	The first time the panel is shown it is populated with the information from the current map area.
		OnShow = function(self, frame)
			self:ProcessInitialUpdate()
			if WhollyDatabase.showsInLogQuestStatus then
				self:ScrollFrame_Update()
			end
		end,

		PrettyNPCString = function(self, npcName, mustKill, realAreaId)
			if mustKill then npcName = format(self.s['MUST_KILL_PIN_FORMAT'], npcName) end
			if realAreaId then npcName = format("%s => %s", npcName, Grail.mapAreaMapping[realAreaId]) end
			return npcName
		end,

		PrettyQuestCountString = function(self, questTable)
			local retval = ""
			if nil ~= questTable then
				local totals = { ['B'] = 0, ['C'] = 0, ['D'] = 0, ['G'] = 0, ['I'] = 0, ['L'] = 0, ['P'] = 0, ['R'] = 0, ['U'] = 0, ['W'] = 0, }
				local code
				for i = 1, #questTable do
					code = questTable[i][2]
					totals[code] = totals[code] + 1
				end
				local colorCode
				for _,code in pairs({ 'G', 'W', 'L', 'P', 'D', 'R', 'I', 'C', 'B', }) do
					colorCode = self.color[code]
					retval = retval .. "|c" .. colorCode .. totals[code] .. "|r/"
				end
				retval = retval .. "|c" .. self.color['U'] .. totals['U'] .. "|r"
			end
			return retval
		end,

		PrettyQuestString = function(self, questTable)
			local questId = questTable[1]
			local filterCode = questTable[2]
			local statusCode = Grail:StatusCode(questId)
			local questLevel = Grail:QuestLevel(questId)
			local questLevelString = WhollyDatabase.prependsQuestLevel and format("[%s] ", questLevel or "??") or ""
			local requiredLevelString = ""
			if WhollyDatabase.appendRequiredLevel then
				local success, _, questLevelNeeded, _ = Grail:MeetsRequirementLevel(questId)
				if bit.band(statusCode, Grail.bitMaskLevelTooLow) > 0 then requiredLevelString = format(" [%s]", questLevelNeeded) end
			end
			local colorCode = self.color[filterCode]
			local repeatableCompletedString = ('D' == filterCode or 'R' == filterCode) and WhollyDatabase.showsAnyPreviousRepeatableCompletions and bit.band(statusCode, Grail.bitMaskEverCompleted) > 0 and "*" or ""
			return format("|c%s%s%s%s%s|r", colorCode, questLevelString, Grail:QuestName(questId), repeatableCompletedString, requiredLevelString)
		end,

		ProcessInitialUpdate = function(self)
			if not self.initialUpdateProcessed then
				self.currentPanelMapID = GetCurrentMapAreaID()
				self:ForcePanelMapArea()
				self.initialUpdateProcessed = true
			end
		end,

		QuestFailureString = function(self, statusBitMask)
			local retval = ""
			if bit.band(statusBitMask, Grail.bitMaskClass + Grail.bitMaskAncestorClass) > 0 then
				retval = retval .. "|TInterface\\MINIMAP\\TRACKING\\Class:0|t"
--				retval = retval .. "|TInterface\\TARGETINGFRAME\\UI-Classes-Circles:0|t" -- contains 10 class circles in one image.  if we decided to show what specific classes were able to get the quest we could display these circles
--				retval = retval .. "|TInterface\\WorldStateFrame\\ICONS-CLASSES:0|t" -- contains 10 class squares in one image.  if we decided to show what specific classes were able to get the quest we could display these circles
			end
			if bit.band(statusBitMask, Grail.bitMaskRace + Grail.bitMaskAncestorRace) > 0 then
--				retval = retval .. "Race"
				retval = retval .. "|TInterface\\BUTTONS\\UI-MicroButton-CharacterNightElf-Up:0|t"
			end
			if bit.band(statusBitMask, Grail.bitMaskGender + Grail.bitMaskAncestorGender) > 0 then
--				retval = retval .. "Gender"
				retval = retval .. format("|TInterface\\CHARACTERFRAME\\TEMPORARYPORTRAIT-%s-%s:0|t", (3 == Grail.playerGender) and "MALE" or "FEMALE", Grail.playerRace:upper())
			end
			if bit.band(statusBitMask, Grail.bitMaskFaction + Grail.bitMaskAncestorFaction) > 0 then
--				retval = retval .. "|TInterface\\MINIMAP\\TRACKING\\BattleMaster:0|t"
				if 'Horde' == Grail.playerFaction then
					retval = retval .. "|TInterface\\BattlefieldFrame\\Battleground-Alliance:0|t"
				else
					retval = retval .. "|TInterface\\BattlefieldFrame\\Battleground-Horde:0|t"
				end
			end
			if bit.band(statusBitMask, Grail.bitMaskInvalidated + Grail.bitMaskAncestorInvalidated) > 0 then
--				retval = retval .. "Invalidated"
				retval = retval .. "|TInterface\\LootFrame\\LootPanel-Icon:0|t"
			end
			if bit.band(statusBitMask, Grail.bitMaskProfession + Grail.bitMaskAncestorProfession) > 0 then
				retval = retval .. "|TInterface\\MINIMAP\\TRACKING\\Profession:0|t"
			end
			if bit.band(statusBitMask, Grail.bitMaskReputation + Grail.bitMaskAncestorReputation) > 0 then
				retval = retval .. "Reputation"
--				retval = retval .. "|TInterface\\PaperDollInfoFrame\\PaperDollSlidebarTabs:3|t" -- this contains a column of images, and we want 3rd or 4th
			end
			if bit.band(statusBitMask, Grail.bitMaskHoliday + Grail.bitMaskAncestorHoliday) > 0 then
				retval = retval .. "|TInterface\\Calendar\\Holidays\\Calendar_HallowsEndStart:0|t"
				-- we could determine the holiday relevance of the quest and put up the specific icon, but they are so small in general it is probably silly
			end
			if bit.band(statusBitMask, Grail.bitMaskLevelTooHigh + Grail.bitMaskAncestorLevelTooHigh) > 0 then
				retval = retval .. "|TInterface\\MINIMAP\\TRACKING\\POIArrow:0|t"
			end
			return retval
		end,

		ScrollFrame_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrame_Update
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonTemplate")
		end,

		ScrollFrame_Update = function(self)
			self = self or Wholly
			self:FilterPanelQuests()
			local questsInMap = self.filteredPanelQuests
			local numEntries = #questsInMap
			local buttons = com_mithrandir_whollyFrameScrollFrame.buttons
			local numButtons = #buttons
			local scrollOffset = HybridScrollFrame_GetOffset(com_mithrandir_whollyFrameScrollFrame)
			local buttonHeight = buttons[1]:GetHeight();
			local button, questIndex, questId, questLevelString, requiredLevelString, colorCode, questLevel, filterCode, repeatableCompletedString
			local shouldShowTag

			table.sort(questsInMap, Wholly.SortingFunction)
			for i = 1, numButtons do
				button = buttons[i]
				questIndex = i + scrollOffset
				if questIndex <= numEntries then
					questId = questsInMap[questIndex][1]
					filterCode = questsInMap[questIndex][2]
					button.normalText:SetText(self:PrettyQuestString(questsInMap[questIndex]))
					shouldShowTag = false
					if 'I' == filterCode and WhollyDatabase.showsInLogQuestStatus and WhollyDatabase.showsQuestsInLog then
						local questStatus = Grail:StatusCode(questId)
						local statusText = nil
						shouldShowTag = true
						if bit.band(questStatus, Grail.bitMaskInLogComplete) > 0 then statusText = COMPLETE
						elseif bit.band(questStatus, Grail.bitMaskInLogFailed) > 0 then statusText = FAILED
						else shouldShowTag = false
						end
						if nil ~= statusText then
							button.tag:SetText("(" .. statusText .. ")")
						end
					end
					if shouldShowTag then button.tag:Show() else button.tag:Hide() end
					button.questId = questId
					button.statusCode = filterCode
					if self.debug then button.reason = questsInMap[questIndex][3] end
					button:Show()
				else
					button:Hide()
				end
			end
			HybridScrollFrame_Update(com_mithrandir_whollyFrameScrollFrame, numEntries * buttonHeight, numButtons * buttonHeight)
		end,

		SetCurrentZone = function(self, frame)
			self:UpdateQuestCaches(false, false, true)
			self:ZoneButtonEnter(frame)	-- need to update the tooltip which is showing
		end,

		ShowBreadcrumbInfo = function(self)
			local questId = GetQuestID()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			com_mithrandir_whollyBreadcrumbFrame:Hide()
			if nil ~= breadcrumbs then
				if 1 == #breadcrumbs then com_mithrandir_whollyBreadcrumbFrameMessage:SetText(self.s['SINGLE_BREADCRUMB_FORMAT'])
				else com_mithrandir_whollyBreadcrumbFrameMessage:SetText(format(self.s['MULTIPLE_BREADCRUMB_FORMAT'], #breadcrumbs))
				end
				com_mithrandir_whollyBreadcrumbFrame:Show()
			end
		end,

		ShowPin = function(self, questTable)
			local codeMapping = { ["G"] = 1, ["W"] = 2, ["D"] = 3, ["R"] = 4, ["P"] = 5, ["L"] = 6, ["U"] = 7 }
			local id = questTable[1]
			local locations = Grail:MultipleUniqueMapLocationQuest(id, 'A', self.currentPinMapID, self.currentPinMapLevel)
			if nil ~= locations then
				for i = 1, #locations do
					local xcoord, ycoord, npcName, npcId = locations[i][1], locations[i][2], locations[i][3], tonumber(locations[i][4])
					if nil ~= xcoord then
						-- record the NPC as needing a tooltip note for the specific quest (it can be a redirect because an actual "NPC" may be the item that starts the quest)
						local shouldProcess, kindsOfNPC = Grail:IsTooltipNPC(npcId)
						if shouldProcess then
							for i = 1, #(kindsOfNPC), 1 do
								local npcIdToUse = npcId
								local shouldAdd = true
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
									shouldAdd = self:DroppedItemMatchesQuest(kindsOfNPC[i][2], id)
								end
								if kindsOfNPC[i][1] == Grail.NPC_TYPE_BY then npcIdToUse = tonumber(kindsOfNPC[i][2]) end
								if nil == self.npcs[npcIdToUse] then self.npcs[npcIdToUse] = {} end
								if shouldAdd and not tContains(self.npcs[npcIdToUse], id) then tinsert(self.npcs[npcIdToUse], id) end
							end
						end

						local pin = self:GetPin(npcId)
						local code = questTable[2]
						if 'D' == code and Grail:IsRepeatable(id) then code = 'R' end
						local codeValue = codeMapping[code]
						local pinValue = codeMapping[pin.texType]
						if codeValue < pinValue then
							pin:SetType(code)
						end						
						pin:ClearAllPoints()
						pin.questId = id
						pin:SetFrameLevel(WorldMapPOIFrame:GetFrameLevel() - 1)
						pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/100*WorldMapDetailFrame:GetWidth(), -ycoord/100*WorldMapDetailFrame:GetHeight())
						pin:Show()

						self.regPins[id..":"..npcId] = pin
					end
				end
			end
		end,

		ShowTooltip = function(self, pin)
			local listedQuests = {}
			self.tooltip:SetOwner(pin, "ANCHOR_RIGHT")
			self.tooltip:ClearLines()
	
			-- find all quests in range of hover
			local mx, my = self:GetMousePosition()
			local npcList = {}
			local npcNames = {}

			local mapId = self.currentPinMapID
			local mapLevel = self.currentPinMapLevel
			local questsInMap = self.filteredPinQuests
			local questId
			for i = 1, #questsInMap do
				questId = questsInMap[i][1]
				local locations, mustKill = Grail:MultipleUniqueMapLocationQuest(questId, 'A', mapId, mapLevel)
				if nil ~= locations then
					for j = 1, #locations do
						local xcoord, ycoord, npcname, npc, areaId, dungeonLevel, realAreaId = locations[j][1], locations[j][2], locations[j][3], locations[j][4], locations[j][5], locations[j][6], locations[j][7]
						if nil ~= xcoord then
							local dist = self:Distance(mx, my, xcoord/100, ycoord/100)
							if (dist <= 0.02) then
								if (not npcList[npc]) then
									npcList[npc] = {}
									npcname = self:PrettyNPCString(npcname, mustKill, realAreaId)
									npcNames[npc] = npcname
								end
								table.insert(npcList[npc], questsInMap[i])
							end
						end
					end
				end
			end

			local first = true;
			for npc,questList in pairs(npcList) do
				if (not first) then
					self.tooltip:AddLine(" ")
				else
					first = false
				end
				for _,qt in ipairs(questList) do
					local leftStr = self:PrettyQuestString(qt)
					local q = qt[1]
					local escortString = Grail:IsEscort(q) and self.s['ESCORT'] or ""
					local dungeonString = Grail:IsDungeon(q) and CALENDAR_TYPE_DUNGEON or ""
					local raidString = Grail:IsRaid(q) and CALENDAR_TYPE_RAID or ""
					local pvpString = Grail:IsPVP(q) and CALENDAR_TYPE_PVP or ""
					local groupString = Grail:IsGroup(q) and CHANNEL_CATEGORY_GROUP or ""
					local heroicString = Grail:IsHeroic(q) and PLAYER_DIFFICULTY2 or ""
					local rightStr = format("%s%s%s%s%s%s", escortString, dungeonString, raidString, pvpString, groupString, heroicString)
					if strlen(rightStr) > 0 then rightStr = format("|c%s%s|r", self.color[qt[2]], rightStr) end

					-- check if already printed - this is for spam quests like the human starting area that haven't been labeled correctly
					if not questName or not listedQuests[questName] then
						self.tooltip:AddDoubleLine(leftStr, rightStr)
						self.tooltip:SetLastFont(self.tooltip.large)
						self.tooltip:SetLastFont(self.tooltip.small, true)
						if questName then listedQuests[questName] = true end
					end
				end
				self.tooltip:AddLine(npcNames[npc], 1, 1, 1, 1)
				self.tooltip:SetLastFont(self.tooltip.small)
			end
	
			self.tooltip:Show();
		end,

		SlashCommand = function(self, frame, msg)
			self:ToggleUI()
		end,

		Sort = function(self)
			-- This is supposed to cycle through the supported sorting techniques and make the contents of the panel
			-- show the quests based on those techniques.
			-- 1 Quest alphabetical
			-- 2 Quest level (and then alphabetical)
			-- 3 Quest type (and then alphabetical)
			WhollyDatabase.currentSortingMode = WhollyDatabase.currentSortingMode + 1
			if (WhollyDatabase.currentSortingMode > 3) then WhollyDatabase.currentSortingMode = 1 end
			self:ScrollFrame_Update()
		end,

		SortingFunction = function(a, b)
			local retval = false
			if 1 == WhollyDatabase.currentSortingMode then
				retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
			elseif 2 == WhollyDatabase.currentSortingMode then
				local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
				if aLevel == bLevel then
					retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
				else
					retval = aLevel < bLevel
				end
			elseif 3 == WhollyDatabase.currentSortingMode then
				local aCode, bCode = a[2], b[2]
				if aCode == bCode then
					retval = Grail:QuestName(a[1]) < Grail:QuestName(b[1])
				else
					retval = aCode < bCode
				end
			end
			return retval
		end,

		ToggleUI = function(self)
			if com_mithrandir_whollyFrame:IsShown() then
				com_mithrandir_whollyFrame:Hide()
			else
				com_mithrandir_whollyFrame:Show()
			end
		end,

		---
		--	Sets up the event monitoring to handle those associated with displaying breadcrumb information.
		UpdateBreadcrumb = function(self)
			if WhollyDatabase.displaysBreadcrumbs then
				com_mithrandir_whollyFrame:RegisterEvent("QUEST_DETAIL")
				if QuestFrame:IsVisible() then
					self:ShowBreadcrumbInfo()
				end
			else
				com_mithrandir_whollyFrame:UnregisterEvent("QUEST_DETAIL")
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end
		end,

		UpdatePins = function(self)
			self:HideAllPins()
			if WhollyDatabase.displaysMapPins then
				self:FilterPinQuests()
				local questsInMap = self.filteredPinQuests
				for i = 1, #questsInMap do
					self:ShowPin(questsInMap[i])
				end
			end
		end,

		UpdateQuestCaches = function(self, forceUpdate, setPinMap, setPanelMap)
			local mapArea = GetCurrentMapAreaID()
			local dungeonLevel = GetCurrentMapDungeonLevel() or 0
			if mapArea ~= self.currentPinMapID or dungeonLevel ~= self.currentPinMapLevel or forceUpdate then
				if setPinMap then self.currentPinMapID = mapArea self.currentPinMapLevel = dungeonLevel end
				self.cachedPinQuests = self:ClassifyQuestsInMap(self.currentPinMapID) or {}
				self:UpdatePins()
			end
			if mapArea ~= self.currentPanelMapID or forceUpdate then
				if setPanelMap then self.currentPanelMapID = mapArea end
				self:ForcePanelMapArea()
			end
		end,

		ZoneButtonEnter = function(self, frame)
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:AddLine(Wholly.panelCountLine)
			GameTooltip:Show()
			GameTooltip:ClearAllPoints()
		end,

		}

	local locale = GetLocale()
	local C = Wholly.configuration
	if "deDE" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Zeige |cFF0099CCwiederholbaren Quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Zeige |cFF00FF00abgeschlossene Quests|r"
		C[6][0]="Zeige |cFF996600unerreichbar Quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Anzeige Pins für Questgeber"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sortieren",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Abgeben:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "esES" == locale or "esMX" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Matar para iniciar [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Matar]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Ordenar",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Regresar:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "frFR" == locale then
		C[1][0]="Ajouter au début du nom de la quête, le niveau de la quête"
		C[2][0]="Append required level"
		C[3][0]="Montrer |cFF0099CCquêtes répétables|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Montrer les |cFF996600quêtes inaccessibles|r"
		C[7][0]="Montrer les |cFF666666quêtes de bas niveau|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Montrer les quêtes de haut niveau"
		C[10][0]="Montrer les |cFFFF0000quêtes qui ont besoin de conditions préalables|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Afficher repères de la carte pour les donneurs de quêtes"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Tuer pour commencer [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Tuer]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Classer",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Rendre:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "koKR" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "완료:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "ptBR" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Entregar:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "ruRU" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Показать |cFF0099CCповторяющиеся квесты|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Убить, чтобы начать [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Убить]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Сортировать",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "Сдать:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "zhCN" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "完成:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	elseif "zhTW" == locale then
		C[1][0]="Prepend quest level"
		C[2][0]="Append required level"
		C[3][0]="Show |cFF0099CCrepeatable quests|r"
		C[4][0]="Show |cFFFF00FFquests in log|r"
		C[5][0]="Show |cFF00FF00completed quests|r"
		C[6][0]="Show |cFF996600unobtainable quests|r"
		C[7][0]="Show |cFF666666low-level quests|r"
		C[8][0]="Show whether repeatable quests previously completed"
		C[9][0]="Show high-level quests"
		C[10][0]="Show |cFFFF0000quests that need prerequisites|r"
		C[11][0]="Quest log panel updates when zones change"
		C[12][0]="Display map pins for quest givers"
		C[13][0]="Display checkbox on world map"
		C[14][0]="Display dungeon quests in outer map"
		C[15][0]="Show prerequisite quests in tooltip"
		C[16][0]="Display breadcrumb quest information on Quest Frame"
		C[17][0]="Display holiday quests always"
		C[18][0]="Open world map updates when zones change"
		C[19][0]="Show status of quests in log"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['QUESTGIVERS'] = "Questgivers:",
			['SORT'] = "Sort",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['TURNINS'] = "完成:",
			['PROFESSION_REQUIRED'] = "Profession Required:",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			}
	end

end
